#include "license.readme"

#include <cstring>
#include <cstdlib>
#include <cstdio>

#include "myspell.hxx"

using namespace std;

extern void mychomp(char * s);
extern char * mystrdup(const char *);


MySpell::MySpell(const char * affpath, const char * dpath)
{
    encoding = NULL;
    csconv = NULL;

    /* first set up the hash manager */
    pHMgr = new HashMgr(dpath);

    /* next set up the affix manager */
    /* it needs access to the hash manager lookup methods */
    pAMgr = new AffixMgr(affpath,pHMgr);

    /* get the preferred try string and the dictionary */
    /* encoding from the Affix Manager for that dictionary */
    char * try_string = pAMgr->get_try_string();
    encoding = pAMgr->get_encoding();
    csconv = get_current_cs(encoding);

    /* and finally set up the suggestion manager */
    pSMgr = new SuggestMgr(try_string, 15, pAMgr);

// ------- missing?  (N.L.) ---------------------------------    
    if (try_string) free(try_string); // NL
// ----------------------------------------------------------
    
    putstate = 0;
    putword = NULL;
    
    prevword=NULL;
}

MySpell::~MySpell()
{
    if (pSMgr) delete pSMgr;
    if (pAMgr) delete pAMgr;
    if (pHMgr) delete pHMgr;
    pSMgr = NULL;
    pAMgr = NULL;
    pHMgr = NULL;
    csconv= NULL;
    if (encoding) free(encoding);
    encoding = NULL;
}


// make a copy of src at destination while removing all leading
// blanks and removing any trailing periods after recording
// their presence with the abbreviation flag
// also since already going through character by character, 
// set the capitalization type
// return the length of the "cleaned" word

int MySpell::cleanword(char * dest, const char * src, int * pcaptype, int * pabbrev)
{ 
   unsigned char * p = (unsigned char *) dest;
   const unsigned char * q = (const unsigned char * ) src;

   // first skip over any leading blanks
   while ((*q != '\0') && (*q == ' ')) q++;
   
   // now strip off any trailing periods (recording their presence)
   *pabbrev = 0;
   int nl = strlen((const char *)q);
   while ((nl > 0) && (*(q+nl-1)=='.')) {
       nl--;
       *pabbrev = 1;
   }
   
   // if no characters are left it can't be an abbreviation and can't be capitalized
   if (nl <= 0) { 
       *pcaptype = NOCAP;
       *pabbrev = 0;
       *p = '\0';
       return 0;
   }

   // now determine the capitalization type of the first nl letters
   int ncap = 0;
   int nneutral = 0;
   int nc = 0;
   while (nl > 0) {
       nc++;
       if (csconv[(*q)].ccase) ncap++;
       if (csconv[(*q)].cupper == csconv[(*q)].clower) nneutral++;
       *p++ = *q++;
       nl--;
   }
   // remember to terminate the destination string
   *p = '\0';

   // now finally set the captype
   if (ncap == 0) {
        *pcaptype = NOCAP;
   } else if ((ncap == 1) && csconv[(unsigned char)(*dest)].ccase) {
        *pcaptype = INITCAP;
  } else if ((ncap == nc) || ((ncap + nneutral) == nc)){
        *pcaptype = ALLCAP;
  } else {
        *pcaptype = HUHCAP;
  }
  return nc;
} 
       

void MySpell::mkallcap(char * p)
{
  while (*p != '\0') { 
    *p = csconv[((unsigned char) *p)].cupper;
    p++;
  }
}


void MySpell::mkallsmall(char * p)
{
  while (*p != '\0') { 
    *p = csconv[((unsigned char) *p)].clower;
    p++;
  }
}


int MySpell::spell(const char * word)
{
  struct hentry * rv=NULL;
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];

  int wl = strlen(word);
  if (wl > (MAXWORDLEN - 1)) return 0;
  
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 1;

  int n = 0;

  // test numbers
  while ((n<wl) && (cw[n]<='9') && (cw[n]>='0')) { n++; }
  if (n==wl) return 1;

  if (prevword && (strcmp(prevword,cw)==0) && (putstate!=2)) return 0;
  free(prevword);
  prevword=mystrdup(word);

  switch(captype) {
     case HUHCAP: {
                    if ((putstate==0) && (strcmp(word,"PuT")==0)) {
                      putstate = 1;
		      return 1;
		    }
                  }
     case NOCAP:   { 
                     rv = check(cw); 
                     if ((abbv) && !(rv)) {
		         memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         rv = check(wspace);
                     }
		     break;
                   }
     case ALLCAP:
     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     rv = check(wspace);
                     if (!rv) rv = check(cw);
                     if ((abbv) && !(rv)) {
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         rv = check(wspace);
			 if (!rv) {
		            memcpy(wspace,cw,wl);
                            *(wspace+wl) = '.';
                            *(wspace+wl+1) = '\0';
			 }
                     }
		     break;
                   }		   
  }
  
  
  // experimental add word function
  // usage: when spelled text contains this command:
  // PuT new_word pattern_word
  // that new_word put into the hashtable 
  // with exactly the same flags of pattern_word
  if (putstate==1) {
    putword = mystrdup(word);
    putstate = 2;
    return 1;
  } else if (putstate==2) {
    if (strcmp(word,putword)==0) {
      putstate=0;
      return pHMgr->putword(putword,strlen(putword),"/",1);
    }
    if (rv==NULL) {
	return 0;
    }
    putstate=0;
    return pHMgr->putword(putword,strlen(putword),rv->astr,rv->alen);
    free(putword);
  } 

  if (rv) return 1;
  
  // compound word with dash (HU)
  char * dash;
  if ((dash=(char *) TESTAFF(cw,'-',wl))) {
      *dash='\0';      
      // hard-coded solution for a hungarian mistake in datum (HU)
      if (strcmp(cw,"1")==0) {
        if ((strcmp(dash+1,"e")==0) ||	// 1-e -> 1-je
	    (strcmp(dash+1,"i")==0) ||	// 1-i -> 1-ji
	    (strcmp(dash+1,"n")==0) ||	// 1-n -> 1-jn
	    (strcmp(dash+1,"ben")==0) ||	// 1-ben -> 1-jben
	    (strncmp(dash+1,"tek",4)==0)) return 0; // 1-tek -> 1-tek
      }  
      // examine 2 sides of the dash
      if (dash[1]=='\0') { // base word ending with dash
	n = spell(cw);
      } else {
      // first word ending with dash: word-
        char r2 = *(dash+1);
        dash[0]='-';
        dash[1]='\0';
        n = spell(cw);
        dash[1] = r2;
        dash[0]='\0';
      }
      if ((n) && spell(dash+1)) return 1;
      // affixed number in correct word
      if (n && (((*(dash-1)<='9') && (*(dash-1)>='0')) || (*(dash-1)>='.'))) {
         *dash='-';
	 n=1;
	 if (*(dash-n)=='.') n++;
	 // search first not a number character to left from dash
	 while (((dash-n)>=cw) && ((*(dash-n)=='0') || (n<3)) && (n<6)) {
	    n++;
	 }
	 if ((dash-n)<cw) n--;
	 // numbers: valami1000000-hoz
	 // examine 100000-hoz, 10000-hoz 1000-hoz, 10-hoz,
	 // 56-hoz, 6-hoz
	 for(; n>=1; n--) {
	    if (check(dash-n)) return 1;
	 }
      }
  }
  return 0;
}


struct hentry * MySpell::check(const char * word)
{
  struct hentry * he = NULL;
  int len;

  forbidden_compound = 0;
  prevcompound=0;
  prevroot=NULL;
    
  if (pHMgr)
     he = pHMgr->lookup(word);

    // check forbidden words
  if ((he) && (he->astr) && (pAMgr) &&
    TESTAFF(he->astr,pAMgr->get_forbiddenword(),he->alen)) {
    	forbidden_compound = 1;
	return NULL;
//	((he->astr)[0]==(pAMgr->get_forbiddenword())))
  }

  len = strlen(word);

  if ((he == NULL) && (pAMgr)) {
     // try stripping off affixes */
     he = pAMgr->affix_check(word, len);

     // try check compound word
     if (he) {
        if ((he->astr) && (pAMgr) &&
	    TESTAFF(he->astr,pAMgr->get_forbiddenword(),he->alen)) {
		forbidden_compound = 1;
		return NULL;
	    }
	prevroot=he->word;
     } else if ((pAMgr->get_compound())!='\0') {
          he = pAMgr->compound_check(word, len, 
	                          pAMgr->get_compound(),0,0,100,0);
	  if ((!he) && (word[len-1]=='-')) {
	     char * dup = mystrdup(word);
	     dup[len-1] = '\0';
             he = pAMgr->compound_check(dup, len-1, 
	                          pAMgr->get_compound(),-5,0,100,1); // hu_mov_rule
	     dup[len-1] = '-';
	     free(dup);
	  }
	  prevcompound=1;
     }


  }
  if (he) return he;
  return NULL;
}

int MySpell::suggest(char*** slst, const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;
  int ns = 0;
  *slst = NULL; // HU, nsug in pSMgr->suggest
  
/*  int n=0; */

  switch(captype) {
     case NOCAP:   { 
                     ns = pSMgr->suggest(slst, cw, ns);
                     if (ns>0) break;
		     
		     /*
		     n=wl;
		     while ((cw[n]<='9') && (cw[n]>='0')) { n++; }
		     if (n==0) break;
		     if (n==wl) {
		         n-=1;
		     } else if (n>5) {
		         n-=4;
		     } else {
		         n=0;
		     }		     
		     do {         // loop for affixed number suggestion (HU)
		         if ((n<wl) && (cw[n]<='9') && (cw[n]>='0')) {
			     n++;
			 } else {
			     break;
			 }
		     } while ((cw[n]<='9') && (cw[n]>='0') && 
		             !(ns=suggest(slst,cw+n)));
                     for (int j=0; j < ns; j++) {
		            memcpy(wspace,cw,n);
			    wspace[n]='\0';
			    strcat(wspace,(*slst)[j]);
			    free((*slst)[j]);
			    (*slst)[j]=mystrdup(wspace);
		     }
		     */
		     break;
                   }

     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest(slst, wspace, ns);
                     for (int j=0; j < ns; j++)
                       mkinitcap((*slst)[j]);

// ------- missing suggestions for proper nouns (N.L.) -----
//                     if (ns == 0) {
	             ns = pSMgr->suggest(slst, cw, ns);
//		     }
// ----------------------------------------------------------
                     break;
		     
                   }

     case HUHCAP: { 
                     ns = pSMgr->suggest(slst, cw, ns);
                     if (ns == 0) {
                        memcpy(wspace,cw,(wl+1));
                        mkallsmall(wspace);
                        ns = pSMgr->suggest(slst, wspace, ns);
		     }
                     break;
                   }

     case ALLCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest(slst, wspace, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
//                     if (ns == 0) {
		         mkinitcap(wspace);
		         ns = pSMgr->suggest(slst, wspace, ns);
//		     }
// ----------------------------------------------------------

                     for (int j=0; j < ns; j++)
                       mkallcap((*slst)[j]);
                     break;
                   }
  }
  
  // replace '-' with ' '
  if (forbidden_compound==1) {
    for (int j=0; j < ns; j++) {
	char * pos = (char *) TESTAFF((*slst)[j],'-',strlen((*slst)[j]));
	if (pos) *pos = ' ';
    }
  }
  return ns;
}


char * MySpell::get_dic_encoding()
{
  return encoding;
}

char * MySpell::get_wordchars()
{
  return pAMgr->get_wordchars();
}

char * MySpell::get_prevroot()
{
  return prevroot;
}

int MySpell::get_prevcompound()
{
  return prevcompound;
}

void MySpell::mkinitcap(char * p)
{
  if (*p != '\0') *p = csconv[((unsigned char)*p)].cupper;
}

struct cs_info * MySpell::get_csconv()
{
  return csconv;
}

